/*
  BFIND.C -- Binary file search utility.

  Copyright  1997, Jim Mischel

  bfind [options] filenames

  Valid options:
    -d searches subdirectories

    -s<string> is a hex string.  For example, to search for carriage return/
    line feed pairs, specify -s0D0A
    You can search for the nul character.
*/
#include <stdio.h>
#include <dir.h>
#include <string.h>
#include <alloc.h>
#include <ctype.h>

#include "dirspan.h"

#define MAX_SEARCH_LEN 100

static int DoSubdirs = 0;
static int SearchStringLength = 0;
static char SearchString[MAX_SEARCH_LEN];


#define INPUT_BUFFER_SIZE 32768u

static char * InputBuffer;

static int ProcessOptionArg
      (
        const char * Arg
      )
{
  if ((Arg[0] != '-') && (Arg[0] != '/'))
  {
    return 0;
  }

  switch (Arg[1])
  {
    case 'd' :
    case 'D' :
      DoSubdirs = 1;
      break;

    case 's' :
    case 'S' :
    {
      int i, len;
      char * s;

      len = strlen (Arg);
      memset (SearchString, 0, sizeof (SearchString));
      if (len % 2)
      {
        fprintf (stdout, "Invalid search string--must be even length\n");
        break;
      }
      s = SearchString;
      for (i = 2; i < len; i++)
      {
        char c;
        if (!isxdigit (Arg[i]))
        {
          fprintf (stdout, "Invalid character in search string--must be hex digit\n");
          return 1;
        }
        // convert to number
        c = (char)toupper (Arg[i]);
        if (isdigit (c))
        {
          c -= (char)'0';
        }
        else
        {
          c -= (char)('A' - 10);
        }
        if (!(i%2))
        {
          *s = (char)(c << 4);
        }
        else
        {
          *s += (char)c;
          s++;
        }
      }
      SearchStringLength = (int)(s-SearchString);
      break;
    }

    default :
      fprintf (stdout, "Invalid option character %c\n", Arg[1]);
      break;
  }
  return 1;
}


static void ProcessFile
      (
        const char * Filename
      )
{
  FILE * f;
  unsigned long CharsRead;
  unsigned int i;
  unsigned long CurPtr;
  unsigned long wPtr;  // value of CurPtr at WorkBuffer[0]
  int iWork;
  int FilenameOutput;

  f = fopen (Filename, "rb");
  if (f == NULL)
  {
    fprintf (stdout, "Error opening file %s\n", Filename);
    return;
  }

  FilenameOutput = 0;
  iWork = 0;
  CurPtr = 0;
  while ((CharsRead = fread (InputBuffer, 1, INPUT_BUFFER_SIZE, f)) != 0)
  {
    for (i = 0; i < CharsRead; i++)
    {
      char c;
      c = InputBuffer[i];

      if (c == SearchString[iWork])
      {
        // character match.  If first character, set pointer
        if (iWork == 0)
        {
          wPtr = CurPtr + i;
        }
        iWork++;
        // if at end of string, output match
        if (iWork == SearchStringLength)
        {
          if (!FilenameOutput)
          {
            fprintf (stdout, "\n%s\n", Filename);
            FilenameOutput = 1;
          }
          fprintf (stdout, "%08lX\n", wPtr);
          // reset pointer...
          {
            char * p;

            p = memchr (&(SearchString[1]), SearchString[0], SearchStringLength-1);
            if (p == NULL)
            {
              iWork = 0;
            }
            else
            {
              iWork = (int)(p - SearchString);
              wPtr += iWork;
            }
          }
        }
      }
      else if (iWork != 0)
      {
        // reset pointer
        char * p;

        p = memchr (&(SearchString[1]), SearchString[0], SearchStringLength-1);
        if (p == NULL)
        {
          iWork = 0;
        }
        else
        {
          int x;
          x = (int)(p - SearchString);
          if (x >= iWork)
          {
            iWork = 0;
          }
          else
          {
            iWork = (int)(p - SearchString);
            wPtr += iWork;
          }
        }
      }
    } // for
    CurPtr += CharsRead;
  } // while
  fclose (f);
}


static void Usage
      (
        void
      )
{
  fprintf (stdout, "BFIND Binary File Search Utility.\n"
                   "Jim Mischel, 08/02/97\n\n");
  fprintf (stdout, "Usage is bfind [options] filenames\n");
  fprintf (stdout, "Valid options:\n"
                   "-d searches subdirectories\n"
                   "-s<string> is a hex string.  For example, to search for \n"
                   "carriage return/line feed pairs, specify -s0D0A\n"
                   "You can search for the nul character.\n");
}

int main
      (
        int argc,
        char * argv[]
      )
{
  int iarg;

  if (argc < 2)
  {
    Usage ();
    return 1;
  }

  // allocate input buffer
  InputBuffer = malloc (INPUT_BUFFER_SIZE);
  if (InputBuffer == NULL)
  {
    fprintf (stdout, "Unable to allocate memory\n");
    return 1;
  }

  for (iarg = 1; iarg < argc; iarg++)
  {
    char * arg;

    arg = argv[iarg];
    if (!ProcessOptionArg (arg))
    {
      if (SearchStringLength == 0)
      {
        fprintf (stdout, "Must define a search string\n");
        return 1;
      }
//      ProcessFileArg (arg);
      dirspanEnumFiles (arg, DoSubdirs, ProcessFile);
    }
  }

  free (InputBuffer);

  return 0;
}

